from setuptools import setup

setup(
    name='StrongMind-id-mapper-client',
    version='1.3.0',
    packages=['mapper'],
    url='https://github.com/StrongMind/id-mapper-client',
    license='',
    author='team-platform',
    author_email='cinnamon@strongmind.com',
    description='Client for Identifier Mapper'
)
